;;; html-to-texinfo.el --- -*- coding: iso-8859-15 -*-
;; Copyright 2010/2013 Vincent Belache
;;
;; Author: Vincent Belache <vincent.b.1@hotmail.fr>
;; Version: $Id: html-to-texinfo.el,v 1.23 2013-07-27 10:59:40 Vincent Exp $
;; Keywords: Texinfo, HTML, conversion
;; X-URL: not distributed yet
;;
;; This software is governed by the CeCILL license under French law and abiding by the rules of distribution
;; of free software.  You can use, modify and/ or redistribute the software under the terms of the CeCILL
;; license as circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
;;
;; As a counterpart to the access to the source code and rights to copy, modify and redistribute granted by
;; the license, users are provided only with a limited warranty and the software's author, the holder of the
;; economic rights, and the successive licensors have only limited liability.
;;
;; In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
;; and/or developing or reproducing the software by the user in light of its specific status of free software,
;; that may mean that it is complicated to manipulate, and that also therefore means that it is reserved for
;; developers and experienced professionals having in-depth computer knowledge. Users are therefore encouraged
;; to load and test the software's suitability as regards their requirements in conditions enabling the
;; security of their systems and/or data to be ensured and, more generally, to use and operate it in the same
;; conditions as regards security.
;;
;; The fact that you are presently reading this means that you have had knowledge of the CeCILL license and
;; that you accept its terms.
;;
;;; Commentary:

;;

;; Put this file into your load-path and the following into your ~/.emacs:
;;   (require 'html-to-texinfo)

;;; Code:

(provide 'html-to-texinfo)
(eval-when-compile
  (require 'cl))
(require 'eieio)
(require 'calc-ext)
(require 'accents-ascii)
(require 'doubly-linked-list)
(require 'compile)

;;;;##########################################################################
;;;;  User Options, Variables
;;;;##########################################################################

(eval-when-compile (defconst html2texi-suspicious-html-tags
  '("meta" "br" "hr" "link" "img" "frame")
"Liste des balises pour lesquelles le HTML ne suit pas une
syntaxe strictement XML. Par exemple `<br>' est utilis au lieu
de `<br/>'."  ))

(eval-when-compile (defconst html2texi-suspicious-html-tags-re
  (regexp-opt html2texi-suspicious-html-tags)))

(defconst html2texi-non-recursive-tags
  '("p" "li"))

(defconst html2texi-hierarchy-list
  '(
    (li (ul ol))
    (tr (table))
    (th (tr))
    (td (tr))
    (dd (dl))
    (dt (dl))  ))

(defconst html2texi-non-recursive-tags-re
  (regexp-opt html2texi-non-recursive-tags))

(defconst html2texi-filepath-re
  "[^ \t\n><:\"][^ \t\n><:]*\\|\"\\(?:[^\\\"]\\|\\\\[^\\\"]\\)+\"\\|\"\\(?:[^\\\"]\\|\\\\[\\\"]\\)+\"")

(defconst html2texi-texi-buffer-local-variables
  '(html2texi-document-information)
    "Liste des variables dclares localement au tampon Texinfo.")

(defconst html2texi-allowed-markup-in-@center
  '(img b i em tt strong dfn code)
  "Liste des balises autorises pour @center.")

(defconst html2texi-@center-max-size 1000)

(eval-when-compile
  (defconst html2texi-attribute-name-re "\\(?:\\sw\\|-\\)+"
    "Expression rationnelle sans groupe de capture reprsentant un identifiant d'attribut de balise"))

(defmacro html2texi-concat-macro (&rest args)
  "Macro utilise pour concatner des constantes connues au
moment de la byte-compilation --- le but c'est que le code
byte-compil tourne plus vite, en effet la concatnation n'a pas
lieu  l'excution mais  la byte-compilation."
  `(concat ,@ args))

(defclass html2texi-simple-markup ()
  ((class-dependant
    :initarg :class-dependant
    :initform nil
    :custom '(repeat
	  (list
	   (regexp :tag "clef")
	   (string :tag "prologue")
	   (string :tag "pilogue")
	   (boolean :tag "conserver les espaces et retours chariot")
	   ))
    :documentation "\
Liste d'association dont les lments sont de la forme
 (CLEF PROLOGUE PILOGUE)

Supposons que l'objet dcrit le traitement de la balise TAG,
alors lorsque le code HTML `<TAG class=\"CLEF\">CONTENU</TAG>'
est rencontr, alors il est convertie en le code Texinfo
`PROLOGUE TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU.")
   (preamble
    :initarg :preamble
    :type string
    :documentation "\
Prologue par dfault  utiliser lorsque la balise TAG auquel
l'objet correspond ne comprend pas d'attribut nomm `class' dont
la valeur est une clef de la liste d'association de la case `:class-depedent'.")
   (postamble
    :initarg :postamble
    :type string
    :documentation "\
pilogue par dfault  utiliser lorsque la balise TAG auquel
l'objet correspond ne comprend pas d'attribut nomm `class' dont
la valeur est une clef de la liste d'association de la case `:class-depedent'.")
   (space-verb
    :initarg :space-verb
    :initform nil
    :type boolean
    :documentation "\
Vrai lorsque les espaces et retours chariot sont  conserver tels quels."
    ))
  :documentation "\
Un object de type `html2texi-simple-markup'
dcrit le traitement d'une balise simple comme par exemple
<code>.")

(defclass html2texi-convert-restrictions ()
  ((current-restrictions
    :initarg :current-restrictions
    :initform nil
    :documentation "Liste des restrictions de conversion
    courantes. Par exemple dans un @ref les @table sont
    interdite.")
   (stack
    :initform nil
    :documentation "Pile pour ajouter des restrictions, et les ter ensuite."))
  :documentation "Contexte des restrictions de conversions courantes.")

(defmethod html2texi-push-restrictions
  ((this html2texi-convert-restrictions) &rest restrictions)
  "Ajoute les restrictions de conversion RESTRICTIONS."
  (let ((current-restrictions (oref this :current-restrictions))
	new-restrictions)
    (dolist (v restrictions)
      (unless (memq v current-restrictions)
	(push v new-restrictions)))
    (oset this :current-restrictions (append new-restrictions
					    (oref this :current-restrictions)))
    (oset this stack (cons new-restrictions (oref this stack)))))

(defmethod html2texi-pop-restrictions
  ((this html2texi-convert-restrictions))
  "Restaure les restrictions d'avant l'appel  `html2texi-push-restrictions'."
  (let* ((new-stack (oref this stack))
	 (latest (pop new-stack))
	 (current (oref this :current-restrictions)))
    (oset this stack new-stack)
    (dotimes (i (length latest))
      (pop current))
    (oset this :current-restrictions current)))

(defmethod html2texi-has-restrictions-p
  ((this html2texi-convert-restrictions) &rest restrictions)
  "Teste si l'une des restrictions dans RESTRICTIONS est en vigueur."
  (let ((current-restrictions (oref this :current-restrictions))
	ret restriction)
    (while restrictions
      (setq restriction (pop restrictions)
	    ret (memq restriction current-restrictions)
	    restrictions (if ret nil restrictions)))
    ret))

(defmacro html2texi-with-restrictions (restrictions &rest body)
  "Effectue BODY avec les restrictions dans la liste RESTRICTIONS."
 `(progn
    (html2texi-push-restrictions html2texi-convert-restrictions-object ,@restrictions)
    (unwind-protect
	(progn ,@body)
      (html2texi-pop-restrictions html2texi-convert-restrictions-object))))

(defclass html2texi-locator-info (doli2-element)
 ((id
   :initarg :id
   :type string
   :documentation "Identifiant d'ancre.")
  (type
   :initarg :type
   :type symbol
   :documentation "Soit `:@anchor', `:@ref',
    `:@node', soit `:@menu-item'."  )
  (position
   :initarg :position
   :type integer
   :documention "Point dans le tampon Texinfo o l'ancre est utilise.")
  (size
   :initarg :size
   :type integer
   :documentation "Nombre de caractres dans le localisateur."))
:documentation "Dtient l'information concernant une ancre
utilise soit par un `@anchor', soit un `@node', soit un
`@ref'.")

(defclass html2texi-locator-listing (doli2-list)
  ((hash-table
    :initarg :hash-table))
  :documentation "Liste de `html2texi-locator-info', c'est  dire
    de pointeurs sur des `@node', `@ref' ou `@anchor', ou des
    entre de `@menu', de sorte  embellir les noms de localisateurs a
    postriori."  )

(defmethod initialize-instance ((this html2texi-locator-listing) &rest fields)
  (call-next-method)
  (oset this :hash-table (make-hash-table)))

(defmethod html2texi-add-locator ((this html2texi-locator-listing) locator-info)
  (let* ((locator-id (oref locator-info :id))
	 (table (oref this :hash-table)))
    (doli2-add-last this locator-info)
    (puthash
     locator-id
     (cons locator-info (gethash locator-id table))
     table)))

(defclass html2texi-node-info ()
  ((up
    :initarg :up
    :type string
    :documentation "Identifiant du nud suprieur.")
   (next
    :initarg :next
    :type string
    :documentation "Identifiant du nud suivant.")
   (prev
    :initarg :prev
    :type string
    :documentation "Identifiant du nud prcdent.")
   (menu-entry-text
    :initarg :menu-entry-text
    :type string
    :documentation "Texte de l'article de @menu pointant vers le nud considr." )
   (in-menu-p
    :initarg :in-menu-p
    :type boolean
    :initform nil
    :documentation "`t' si le node est dj rfrenc dans un menu, `nil' sinon.")
   (locator
    :initarg :locator
    :documentation "Localisateur pour le nud que cet objet dcrit."))
  :documentation "Objet dcriant les liaison d'un nud avec les autres nuds.")


(defclass html2texi-files-to-do-listing ()
  ((already-to-do
   :initarg :already-to-do
   :initform nil
   :documentation "Liste des fichiers qui ont t trouvs comme
   tant  traiter lors du traitement d'un fichier qui a dj t
   compltement trait.")
   (doing-or-done
    :initarg :doing-or-done
    :initform nil
    :documentation "Liste des fichiers qui ont dj t traits, le
    premier de la liste est le fichier en cours de traitement."
    )
   (added-file-count
    :initarg :added-file-count
    :initform 0
    :documentation "Nombre de fichier qui est t ajout  la
  liste des fichiers  traiter.")
   (soon-to-do
   :initarg :soon-to-do
   :initform nil
   :documentation "Liste des fichiers qui sont trouvs comme
   tant  traiter lors du traitement du fichier en cours de
   traitement.")
   (in-menu-table
    :initarg :in-menu-table
    :documentation "Table de hashage contenant tous les fichiers
    qui sont rfrencs dans un `@menu'. La clef d'indexation est
    le chemin complet du fichier. La valeur d'indexe est un
    `html2texi-node-info'."))
  :documentation "Objet servant  lister les fichiers 
  traiter. Il comprend deux listes: `already-to-do' et
  `soon-to-do' parce que lors du traitement d'un fichier TOTO les
  nouveaux fichiers  traiter sont mis dans `soon-to-do' dans
  l'ordre o ils sont rencontrs, du coup une fois que le fichier
  TOTO a compltement t trait, on inverse cet ordre en
  transvasant le contenu de `soon-to-do' dans `already-to-do'.")

(defclass html2texi-table-fmt-ctxt ()
  ((col-number
    :initarg :col-number
    :initform 0
    :type integer)
   (row-number
    :initarg :row-number
    :initform 0
    :type integer)
   (head-on-row-0
    :initarg :head-on-row-0
    :initform nil
    :type boolean)
   (force-head
    :initarg :force-head
    :initform nil
    :type boolean)
   (col-count
    :initarg :col-count
    :initform 0
    :type integer
    :documentation
    "Nombre de colonnes dans le tableau.")
   (col-info
    :initarg :col-info
    :documentation
    "Liste d'information sur chaque colonne. Le premier lment
    est factice est ne correspond  aucune colonne.")
   (col-info-last
    :initarg :col-info-last
    :documentation
    "Pointe sur la dernire cons-cell de l'attribut `:col-info'.")
   (col-info-length
    :initarg :col-info-length
    :initform 0
    :type integer))
  :documentation "Contexte de formattage d'une table.")

(defmethod initialize-instance ((this html2texi-table-fmt-ctxt) &rest fields)
  (call-next-method)
  (let ((ci (list 0)))
    ;; le premier lment ne correspond pas  une colonne mais servira 
    ;; reduire le vecteur des informations sur chaque colonne
    (oset this :col-info ci)
    (oset this :col-info-last ci)))


(defmethod initialize-instance ((this html2texi-files-to-do-listing) &rest fields)
  (call-next-method)
  (oset this :in-menu-table  (make-hash-table)))

(defmethod html2texi-files-make-menu-maybe ((this html2texi-files-to-do-listing))
  (let* ((soon-to-do (oref this :soon-to-do))
	(in-menu-table (oref this :in-menu-table))
	(up (html2texi-current-file this))
	prev 
	menu-files)
    (dolist (f soon-to-do)
      (let ((node-info (gethash f in-menu-table)))
	(when (and (null  (and node-info
			       (oref node-info :in-menu-p)))
		   (file-exists-p f))
	  (unless node-info
	    (puthash f 
		     (make-instance 
		      'html2texi-node-info
		      :locator (concat "file:" f)) 
		     in-menu-table))
	  (push f menu-files))))
    (when menu-files
      (insert "\n@menu\n")
      (while menu-files
	(let* ((f (pop menu-files))
	       (next (and (cdr menu-files) (cadr menu-files)))
	       (node-info (gethash f in-menu-table))
	       (menu-entry-text (if (and node-info (slot-boundp node-info :menu-entry-text))
				    (oref  node-info :menu-entry-text) "")))
	  (html2texi-process-url (concat "file:" f) menu-entry-text :@menu-item)
	  (oset node-info :up up)
	  (when next
	    (oset node-info :next next))
	  (when prev
	    (oset node-info :prev prev))
	  (setq prev f)))
      (insert "@end menu\n"))))

(defmethod html2texi-files-has-to-do ((this html2texi-files-to-do-listing))
  (or (oref this :already-to-do)
      (oref this :soon-to-do))  )

(defmethod html2texi-current-file ((this html2texi-files-to-do-listing))
  (car (oref this :doing-or-done)))

(defmethod html2texi-get-next-file-next-to-do  ((this html2texi-files-to-do-listing))
  "Rcupre le prochain fichier  traiter. L'appel de cette
mthode si aucun fichier n'est  traiter gnre une erreur."
  (let ((soon-to-do (oref this :soon-to-do))
	(already-to-do (oref this :already-to-do)))
    (when soon-to-do
      (while soon-to-do (push (pop soon-to-do) already-to-do))
      (oset this :soon-to-do nil))
    (let ((next (pop already-to-do)))
      (oset this :already-to-do already-to-do)
      (oset this :doing-or-done (cons next (oref this :doing-or-done)))
      next)))

(defmethod html2texi-add-file-to-do   ((this html2texi-files-to-do-listing) next)
  "Ajoute le fichier dont le nom absolu est NEXT  la liste des
fichier  traiter. Renvoie `nil' si le fichier tait dj
connu, non-`nil' sinon."
  (unless (or (member next (oref this :already-to-do))
	      (member next (oref this :doing-or-done))
	      (member next (oref this :soon-to-do)))
    (oset this :soon-to-do (cons next (oref this :soon-to-do)))
    (oset this :added-file-count (1+ (oref this :added-file-count)))))

(defmethod html2texi-add-menu-reference  ((this html2texi-files-to-do-listing) file locator &optional in-menu-p)
  (let ((in-menu-table (oref this :in-menu-table)))
    (if (gethash file in-menu-table)
	(html2texi-fatal-error "Double rfrence au fichier `%s'" :html2texi-generic-error file)
      (puthash file (make-instance 'html2texi-node-info
				   :locator locator
				   :in-menu-p in-menu-p) in-menu-table))))

(defclass html2texi-document-information ()
  ((author
    :initarg :author
    :type string)
   (title
    :initarg :title
    :type string)
   (language
    :initarg :language
    :type string)
   (encoding
    :initarg :encoding
    :type string)
   (header-size
    :initarg :header-size
    :type integer)
   )
  :documentation "Information sur le document Texinfo")

(defun html2texi-texinfo-inside-comment-p ()
  "Renvoie non nil lorsque le point est dans un commentaire Texinfo."
  (save-match-data
    (save-excursion
      (let ((cur (point))
	    (end (progn (end-of-line) (point))))
	(beginning-of-line)
	(and (re-search-forward "\\(^\\|[^@]\\)@c\\(omment\\)\\_>" end t)
	     (<= (match-beginning 0) cur))))))

(defmethod html2texi-handle-simple-markup
  ((this html2texi-simple-markup) xml-expr)
  "\
Traitement d'un balisage simple en insrant le contenue entre un
prologue et un pilogue. Si la l'attribut `class' a une valeur,
et que cette valeur est trouv dans CLASS-ALIST alors le prologue
et l'pilogue en sont dduit par association, sinon PREAMBLE et
POSTAMBLE sont respectivement utilis."
  (let* ((class-attribute (assq 'class (nth 1 xml-expr)))
	 (pre-post (or (cdr-safe
			(and class-attribute
			     (assoc-string (cdr class-attribute)
					   (oref this :class-dependant))))
		      (list (oref this :preamble)
			    (oref this :postamble)))))
    (insert (car pre-post))
    (let ((beg (point))
	  end)
      (html2texi-process-xml-expr xml-expr)
      (unless (oref this :space-verb)
	(setq end (point-marker))
	(goto-char beg)
	(while (re-search-forward "[\n\r]\\s-*" nil end)
	  (let ((replace-str " "))
	    (save-match-data
	      (cond
	       ((html2texi-texinfo-inside-comment-p)
		(setq replace-str nil))))
	      (and replace-str (replace-match replace-str t t)))))
	(goto-char end)
	(set-marker end nil))
    (insert (cadr pre-post))))


(defcustom html2texi-save-texi-buffer-confirm-p
  t
  "`nil' pour sauvegarder sans confirmation le tampon Texinfo
aprs la conversion, `t' sinon."
  :type '(radio
	  (const :tag "Sans confirmation" nil)
	  (const :tag "Demander confirmation" t))
  :group 'html2texi)

(defcustom html2texi-code-simple-markup
  (html2texi-simple-markup "html2texi-code-simple-markup"
   :class-dependant '(("file" "@file{" "}"))
   :preamble "@code{"
   :postamble "}")
  "Liste d'association dont les lments sont de la forme (CLEF PROLOGUE PILOGUE).

Lorsque le code HTML `<code class=\"CLEF\">CONTENU</code>' est
rencontr, alors il est convertie en le code Texinfo `PROLOGUE
TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU."
  :type 'object
  :group 'html2texi)

(defcustom html2texi-reuse-log-buffer t
  "Mettre  `nil' pour que le tampon de sortie des erreurs &
avertissement soit re-gnr avec un nom unique  chaque
traitement."
  :type  '(radio
	  (const :tag "Crer un nouveau tampon d'erreurs  chaque conversion." nil)
	  (const :tag "Rutiliser le tampon d'erreurs s'il existe dj." t))
  :group 'html2texi)

(defcustom html2texi-url-encoding
  :html2texi-utf-8
  "Slectionne le codage des URL."
  :type '(radio
	  (symbol :tag "UTF-8" :html2texi-utf-8)
	  (symbol :tag "ISO-8859-1" :html2texi-latin-1))
  :group 'html2texi)

(defcustom html2texi-i-simple-markup
  (html2texi-simple-markup "html2texi-i-simple-markup"
   :preamble "@i{"
   :postamble "}")
  "Liste d'association dont les lments sont de la forme (CLEF PROLOGUE PILOGUE).

Lorsque le code HTML `<code class=\"CLEF\">CONTENU</code>' est
rencontr, alors il est convertie en le code Texinfo `PROLOGUE
TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU."
  :type 'object
  :group 'html2texi)

(defcustom html2texi-b-simple-markup
  (html2texi-simple-markup "html2texi-b-simple-markup"
   :preamble "@b{"
   :postamble "}")
  "Liste d'association dont les lments (CLEF PROLOGUE PILOGUE) comprend.

Lorsque le code HTML `<b class=\"CLEF\">CONTENU</b>' est
rencontr, alors il est convertie en le code Texinfo `PROLOGUE
TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU."
  :type 'object
  :group 'html2texi)

(defcustom html2texi-code-simple-markup
  (html2texi-simple-markup "html2texi-code-simple-markup"
   :class-dependant '(("file" "@file{" "}"))
   :preamble "@code{"
   :postamble "}")
  "Liste d'association dont les lments (CLEF PROLOGUE PILOGUE) comprend.

Lorsque le code HTML `<code class=\"CLEF\">CONTENU</code>' est
rencontr, alors il est convertie en le code Texinfo `PROLOGUE
TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU."
  :type 'object
  :group 'html2texi)

(defcustom html2texi-kbd-simple-markup
  (html2texi-simple-markup "html2texi-code-simple-markup"
   :preamble "@kbd{"
   :postamble "}")
  "Liste d'association dont les lments (CLEF PROLOGUE PILOGUE) comprend.

Lorsque le code HTML `<kbd class=\"CLEF\">CONTENU</kbd>' est
rencontr, alors il est convertie en le code Texinfo `PROLOGUE
TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU."
  :type 'object
  :group 'html2texi)

(defcustom html2texi-dfn-simple-markup
  (html2texi-simple-markup "html2texi-dfn-simple-markup"
   :preamble "@dfn{"
   :postamble "}")
  "Liste d'association dont les lments (CLEF PROLOGUE PILOGUE) comprend.

Lorsque le code HTML `<kbd class=\"CLEF\">CONTENU</kbd>' est
rencontr, alors il est convertie en le code Texinfo `PROLOGUE
TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU."
  :type 'object
  :group 'html2texi)

(defcustom html2texi-em-simple-markup
  (html2texi-simple-markup "html2texi-em-simple-markup"
   :preamble "@emph{"
   :postamble "}")
  "Liste d'association dont les lments sont de la forme (CLEF PROLOGUE PILOGUE).

Lorsque le code HTML `<code class=\"CLEF\">CONTENU</code>' est
rencontr, alors il est convertie en le code Texinfo `PROLOGUE
TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU."
  :type 'object
  :group 'html2texi)

(defcustom html2texi-sub-simple-markup
  (html2texi-simple-markup "html2texi-sub-simple-markup"
   :preamble ""
   :postamble "")
  "Liste d'association dont les lments (CLEF PROLOGUE PILOGUE) comprend.

Lorsque le code HTML `<sub class=\"CLEF\">CONTENU</sub>' est
rencontr, alors il est convertie en le code Texinfo `PROLOGUE
TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU."
  :type 'object
  :group 'html2texi)

(defcustom html2texi-sup-simple-markup
  (html2texi-simple-markup "html2texi-sup-simple-markup"
   :preamble ""
   :postamble "")
  "Liste d'association dont les lments (CLEF PROLOGUE PILOGUE) comprend.

Lorsque le code HTML `<sup class=\"CLEF\">CONTENU</sup>' est
rencontr, alors il est convertie en le code Texinfo `PROLOGUE
TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU."
  :type 'object
  :group 'html2texi)

(defcustom html2texi-samp-simple-markup
  (html2texi-simple-markup "html2texi-samp-simple-markup"
   :preamble "@samp{"
   :postamble "}")
  "Liste d'association dont les lments sont de la forme (CLEF PROLOGUE PILOGUE).

Lorsque le code HTML `<samp class=\"CLEF\">CONTENU</code>' est
rencontr, alors il est convertie en le code Texinfo `PROLOGUE
TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU."
  :type 'object
  :group 'html2texi)

(defcustom html2texi-strong-simple-markup
  (html2texi-simple-markup "html2texi-strong-simple-markup"
   :preamble "@strong{"
   :postamble "}")
  "Liste d'association dont les lments (CLEF PROLOGUE PILOGUE) comprend.

Lorsque le code HTML `<b class=\"CLEF\">CONTENU</b>' est
rencontr, alors il est convertie en le code Texinfo `PROLOGUE
TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU."
  :type 'object
  :group 'html2texi)


(defcustom html2texi-tt-simple-markup
  (html2texi-simple-markup "html2texi-tt-simple-markup"
   :preamble "@t{"
   :postamble "}")
  "Liste d'association dont les lments sont de la forme (CLEF PROLOGUE PILOGUE).

Lorsque le code HTML `<tt class=\"CLEF\">CONTENU</tt>' est
rencontr, alors il est convertie en le code Texinfo `PROLOGUE
TEXI-CONTENU PILOGUE', o TEXI-CONTENU est obtenu en
convertissant en Texinfo CONTENU."
  :type 'object
  :group 'html2texi)


(defcustom html2texi-handle-two-columns-table-as-@table
  t
 "Si `nil' alors une table `<table>...</table>'
avec deux colonne sera gre en texinfo par une `@table', si non
`nil', alors elle sera gre par une `@multitable'."
  :type
  '(choice
    (const :tag "t pour @table" t)
    (const :tag "nil pour @multitable" nil))
  :group 'html2texi)

(defcustom html2texi-log-error-names
   ["Erreur fatale" "Erreur" "Avertissement" "Info"]
   "Liste des types d'erreur prfixant les messages d'erreur dans
le tampon de sortie des erreurs & avertissement de traitement."
   :type
   '(vector
     (string :tag "Erreur fatale")
     (string :tag "Erreur")
     (string :tag "Avertissement")
     (string :tag "Info"))
   :group 'html2texi  )

(defcustom html2texi-log-buffer-name
  "*HTML2TEXI*"
  "Nom du tampon de sortie des erreurs et avertissements de traitement."
  :type 'string
  :group 'html2texi)

(defcustom html2texi-beautify-locators
  t
  "Embellit les identificateur de localisateur aprs gnration
  du Texinfo."
  :type '(radio
	  (const :tag "Ne pas embellir les identificateurs de localisateur." nil)
	  (const :tag "Embellir les identificateurs de localisateur." t))
  :group 'html2texi )

(defvar html2texi-document-information-object nil
  "Objet de classe `html2texi-document-information' pour mmoriser les informations (titre,
  auteurs, etc...) propres  un document.")

(defvar html2texi-line-delta 0
  "Dcalage entre le numro de ligne du code XML au sein le
  tampon Texinfo en cours de traitement, et son numro de ligne
  dans le fichier HTML source.")

(defvar html2texi-xml-stack nil
  "Pile des expressions XML")

(defvar html2texi-keep-empty-strings nil
  "Non nil si les chanes vides sont  conserver.")

(defvar html2texi-ignore-head nil
  "Non nil si on ignore le <head> (dans un fichier HTML li).")

(defvar html2texi-directory-stack nil
  "Pile des chemins de rpertoire.")

(defvar html2texi-files-to-do nil
  "Base des fichiers non encore traits, instancie localement
  comme un objet de class `html2texi-files-to-do-listing'.")

(defvar html2texi-flushable-anchors nil
  "Liste de nom d'ancrage de lien dont l'insertion a t remise 
  plus tard." )

(defvar html2texi-postpone-output nil
  "Non `nil' lorsque l'insertion du code est remise  plus tard.")

(defvar html2texi-directory-ref nil
  "Rpertoire de rfrence")

(defvar html2texi-log-buffer nil
  "Tampon de sortie des erreurs et avertissements de traitement.")

(defvar html2texi-reusable-log-buffer nil
  "Quand `html2texi-reuse-log-buffer' vaut est non `nil', tampon
  qu'on essaie de reutiliser pour la sortie des erreurs..")

(defvar html2texi-texi-buffer-name nil
  "Nom du tampon Texinfo gnr.")

(defvar html2texi-locator-list nil
  "Liste des localisateurs, pour post-traitement d'embellissement
  des identificateurs de localisateur.")

(defvar html2texi-convert-restrictions-object nil
  "Objet de classe `html2texi-convert-restrictions' comprenant la
  liste courante des restrictions de conversion."  )

(defmacro html2texi-make-simple-markup-handler (tag)
  `(defun ,(intern (concat "html2texi-tag-handler-" (symbol-name tag)))
     (xml-expr)
     (html2texi-handle-simple-markup
      ,(intern (concat "html2texi-"  (symbol-name tag) "-simple-markup"))
      xml-expr) ))

(defun html2texi-make-html-clean-xml (beg end)
  (let  ((end-arg end)
	 end)
    ;; initialisation de end comme un marque
    (if (markerp end-arg)
	(setq end end-arg)
      (goto-char end-arg)
      (setq end (point-marker)))

    ;; traite les scripts pour qu'ils soient sans <![CDATA[ ... ]]>
    (goto-char beg)
    (let ((case-fold-search t) p1 p2 p3 p4 p40)
      (while (re-search-forward "<script\\>[^>]*>" end t)
	(setq p1 (match-end 0))
	(unless (re-search-forward "</script>" end t)
	  (html2texi-fatal-error "Clture non trouve pour la balise <script...>" :html2texi-generic-error ))
	(setq p2 (match-beginning 0))
	(goto-char p1)
	(while (re-search-forward "<!\\[CDATA\\[" p2 t)
	  (setq p3 (match-beginning 0))
	  (unless (re-search-forward "\\]\\]>" p2 t)
	    (html2texi-fatal-error "Balisage <![CDATA[ ... ]]> incomplet" :html2texi-generic-error))
	  (setq p4 (match-beginning 0))
	  (delete-region p4 (+ p4 3)); suppression ]]>
	  (delete-region p3 (+ p3 9)); suppression <!CDATA[
	  (setq p40  (- p4 9); comptage suppression <!CDATA[
		p2 (- p2 3)); comptage suppression  ]]>
	  (goto-char p3)
	  (while (re-search-forward "[<>&]" p40 t)
	    (cond
	     ((string= (match-string 0) "<")
	      (replace-match "&lt;" t t)
	      (setq p40 (+ p40 3)))
	     ((string= (match-string 0) ">")
	      (replace-match "&gt;" t t)
	      (setq p40 (+ p40 3)))
	     ((string= (match-string 0) "&")
	      (replace-match "&amp;" t t)
	      (setq p40 (+ p40 4)))
	     (t (html2texi-fatal-error "Erreur interne" :html2texi-generic-error))))
	  (setq p2 (+ p2 (- p40 p4))))
	  (goto-char (+ p2 9); +9 = longueur de </script>
		     ) ))

    ;; rend les balises implicitement auto-closante vraiment auto-closante
    (goto-char beg)
    (while (re-search-forward (html2texi-concat-macro "<\\(" html2texi-suspicious-html-tags-re "\\)\\>")
			      end t)
      (let ((tag (match-string-no-properties 1)))
	(unless (re-search-forward ">" nil t)
	  (html2texi-fatal-error "Clture non trouv pour la balise %s" :html2texi-generic-error
				 tag))
	(when (null (looking-back "/>"))
	  (backward-char)
	  (insert "/")
	  (forward-char))))

    ;; marque de paragraphe et de ligne
    (goto-char beg)
    (let (tag-stack pos-< pos->
		    tag is-closure self-closing)
      (while (re-search-forward "<\\(/\\)?\\([a-zA-Z]+\\)\\>" end t)
	(setq pos-< (match-beginning 0)
	      tag (match-string-no-properties 2)
	      is-closure (match-string-no-properties 1))
	(unless (string= tag (downcase tag))
	  (replace-match (setq tag (downcase tag)) t t nil 2))
	(unless (re-search-forward ">" end t)
	  (html2texi-fatal-error "Soufflet de clture non trouv pour la balise %s" :html2texi-generic-error
		 tag))

	(setq pos-> (point)
	      self-closing (looking-back "/>"))
	(cond
	 ((and self-closing  is-closure)
	  (html2texi-fatal-error "balise %s  la fois de clture et auto-closante" :html2texi-generic-error
		 tag))

	 (self-closing
	  ;; do nothing
	  )

	 ((null is-closure)
	  (when (and
		 (string-match (concat "\\`" html2texi-non-recursive-tags-re "\\'") tag)
		 tag-stack
		 (string= tag (caar tag-stack)))
	    ;; clture
	    (save-excursion
	      (goto-char pos-<)
	      (insert "</" tag "><!-- HTML2TEXI: repaired (1) -->")
	      (html2texi-warning "Ajout clture `</%s>'" nil tag))
	    (pop tag-stack))
	  (push (list tag) tag-stack))

	 (is-closure
	  (if (and tag-stack
		   (string= (caar tag-stack) tag))
	      (pop tag-stack)
	    (let ((first (assoc-string tag tag-stack)) rev)
	      (if first
		  (progn
		    (while (and tag-stack (null (eq (car tag-stack) first)))
		      (push (pop tag-stack) rev))
		    (pop tag-stack)
		    (save-excursion
		      (goto-char pos-<)
		      (dolist (c rev)
			(insert  "</" (car c) "><!-- HTML2TEXI: repaired (2) -->" )
			(html2texi-warning "Ajout clture `</%s>'" nil tag))))

		(save-excursion
		  (goto-char pos->)
		  (insert "-->")
		  (goto-char pos-<)
		  (insert "<!-- HTML2TEXI: repaired (3). "))
		(html2texi-warning "Clture de %s ne correspondant  aucune ouverture" nil tag)))))))


      (let (p-pos li-pos markup poss pos)
	(while (re-search-forward "<\\(/\\)?\\(p\\|li\\)\\>[^>]*>" end t)
	  (setq markup (match-string-no-properties 2)
		poss (intern (concat markup "-pos"))
		pos (symbol-value poss))
	  (if (match-string-no-properties 1)
	      (if  pos
		  (if (= pos (match-beginning 0))
		      (set poss nil)
		    (html2texi-fatal-error "Clture de balise %s ne correspondant  aucune ouverture" 
					   :html2texi-generic-error
					   (match-string-no-properties 0)))
		(save-excursion
		  (save-match-data
		    (unless  (re-search-forward
			      (format "<\\(/\\)?\\(%s\\)\\>[^>]+>" markup)
			      end t)
		      (html2texi-fatal-error "Ouverture de balise <%s> sans clture" 
					     :html2texi-generic-error
					     markup))
		    (set poss (match-beginning 0))
		    (unless (match-string-no-properties 1)
		      (goto-char (match-beginning 0))
		      (insert (format "</%s>" markup))
		      (dolist (symb '(p-pos li-pos))
			(unless (or (null (symbol-value symb))
				    (eq symb poss))
			  (set symb (+ (length markup) 3 (symbol-value symb)))))))))))
	))

    ;; mise en bas de casse des balises
    (goto-char beg)
    (while (re-search-forward "</?\\([A-Za-z_]*[A-Z][A-Za-z0-9_]*\\)" nil end)
      (replace-match (downcase (match-string 1)) t t nil 1))

    ;; mise entre guillemets anglais des attributs entiers ou symboliques sans guillemet, ou entre apostrophes
    (goto-char beg)
    (let (p1 p2 added-dquote-attributes p-start p-end simple-quote-len)
      ;; tant qu'on pense trouver potentiellement un attribut sans "..."  --- le
      ;; test est heuristique, a ne veut pas dire qu'on a trouv un tel
      ;; attribut, en effet le texte suppos fautif peut faire parti de la
      ;; valeur d'un attribut, ou du composant textuel d'un noeud HTML.
      (while (re-search-forward
	      (html2texi-concat-macro
	       html2texi-attribute-name-re
	       "\\(?:\\s-\\|\n\\)*=\\(?:\\s-\\|\n\\)*\\(\\('?\\)\\(?:[-+]?[0-9]+\\|[A-Za-z_][A-Za-z_0-9]*\\)\\2\\)")
				nil end)
	(setq p1 (match-beginning 1)
	      p2 (match-end 0))
	(if
	    ;; on vrifie si la position est bien entre < et > car a limine le
	    ;; cas ou le texte suppos fautif fait partie du composant textuel
	    ;; d'un noeud HTML.
	    (and (re-search-backward "[<>]" beg t)
		 (looking-at (html2texi-concat-macro "<" html2texi-attribute-name-re))
		 (progn
		   (goto-char (setq p-start (match-end 0)))
		   (re-search-forward "[<>]" end t)
		   (string= (match-string-no-properties 0) ">"))
		 (>= (setq p-end (match-beginning 0)) p1))
	    ;; alors on fait la rparation --- attention il se peut qu'il n'y
	    ;; ait pas de problme dans le cas o le texte suppos fautif ferait
	    ;; partie de la valeur d'un attribut.
	    (progn
	      (goto-char p-start)
	      (setq added-dquote-attributes nil)
	      (while
		  (progn
		    (while (looking-at 
			    (html2texi-concat-macro
			     "\\(?:\\s-\\|\n\\)*"
			     html2texi-attribute-name-re
			     "\\(?:\\s-\\|\n\\)*=\\(?:\\s-\\|\n\\)*\"[^\"]*\""))
		      (goto-char (match-end 0)))
		    (cond
		     ((looking-at
		       (html2texi-concat-macro
			"\\(?:\\s-\\|\n\\)*\\("
			html2texi-attribute-name-re
			"\\)\\(?:\\s-\\|\n\\)*=\\(?:\\s-\\|\n\\)*\\(\\('?\\)\\(?:[-+]?[0-9]+\\|[A-Za-z_][A-Za-z_0-9]*\\)\\3\\)"))
		      (push (match-string-no-properties 1) added-dquote-attributes)
		      (setq simple-quote-len (length (match-string 3)))
		      (goto-char (match-beginning 2))
		      (or (= simple-quote-len 0)
			  (delete-char 1))
		      (insert ?\")
		      (goto-char (+ (- 1 (* 2 simple-quote-len)) (match-end 2)))
		      (insert ?\")
		      (or (= simple-quote-len 0)
			  (delete-char 1))
		      (setq p-end (+ (* 2 (- 1 simple-quote-len)) p-end)))
		     ((looking-at "\\(?:\\s-\\|\n\\)*/?>")
		      (goto-char (match-end 0))
		      (insert "<!-- HTML2TEXI inserted double quotes around values for attibutes: "
			      (mapconcat (lambda (x) (concat "`" x "'"))
					 (nreverse added-dquote-attributes)
					 ", ")
			      " -->")
		      nil)
		     (t 
		      (html2texi-error "Attribut au format invalide: %s."
					 :html2texi-generic-error
					 (buffer-substring (point) p-end)))))))
	  ;; sinon on continue  chercher un attribut potentiel dont la valeur
	  ;; n'est pas entre "..."
	  (goto-char p2))))

    ;; un peu de mnage...
    (unless (markerp end-arg)
      (set-marker end nil)) ))

;;;==========================================================================
;;; dfinition des gestionnaires de balise
;;;--------------------------------------------------------------------------

(defun html2texi-tag-handler-a (xml-expr)
  (let (name href text (xml-expr-length (length xml-expr)))
    (dolist (attrib (cadr xml-expr))
      (cond
       ((eq (car attrib) 'href)
	(setq href (cdr attrib)))
       ((eq (car attrib) 'name)
	(setq name (cdr attrib)))))
    (and (cddr xml-expr)
	 (setq text (if (cdddr xml-expr)
			xml-expr
		      (caddr xml-expr))))
    (cond
     (href (html2texi-process-url href text))
     ((= xml-expr-length 3)
      (cond
       ((stringp text)
	(insert (html2texi-string-escape text t)))
       ((consp text)
	(html2texi-process-xml-expr text)	)
       (t (error "Le format du text de la balise <a> tait inattendu"))))
     ((> xml-expr-length 3)
      (html2texi-process-xml-expr `(div nil ,@(cddr xml-expr)))))
    (and
     name
     (progn
       (insert "\n@anchor{")
       (html2texi-insert-locator
	 (concat
	  (file-relative-name
	   (html2texi-current-file html2texi-files-to-do)
	   html2texi-directory-ref)
	  "#" name) :@anchor)
       (insert "}\n")))))

(html2texi-make-simple-markup-handler b)

(defun html2texi-flush-anchors ()
  (while html2texi-flushable-anchors
    (insert "@anchor{")
    (html2texi-insert-locator (pop html2texi-flushable-anchors) :@anchor)
    (insert "}\n")))

(defun html2texi-tag-handler-h1 (xml-expr)
  (insert "@chapter " (hmtl2texi-to-plain-text xml-expr :one-line) ?\n)
  (html2texi-flush-anchors))


(defun html2texi-tag-handler-h2 (xml-expr)
  (insert "@section " (hmtl2texi-to-plain-text xml-expr :one-line) ?\n)
  (html2texi-flush-anchors))

(defun html2texi-tag-handler-h3 (xml-expr)
  (insert "@subsection " (hmtl2texi-to-plain-text xml-expr :one-line) ?\n)
    (html2texi-flush-anchors))

(defun html2texi-tag-handler-h4 (xml-expr)
  (insert "@subsubsection " (hmtl2texi-to-plain-text xml-expr :one-line) ?\n)
  (html2texi-flush-anchors))

(defun html2texi-tag-handler-h5 (xml-expr)
   (insert "@subsubsection " (hmtl2texi-to-plain-text xml-expr :one-line) ?\n)
     (html2texi-flush-anchors))

(defun html2texi-tag-handler-h6 (xml-expr)
   (insert "@subsubsection " (hmtl2texi-to-plain-text xml-expr :one-line) ?\n)
   (html2texi-flush-anchors))

(html2texi-make-simple-markup-handler samp)

(defun html2texi-get-col-span (xml-expr)
  ;; xml-expr is <td> or <th>
  (let ((col-span (cdr-safe (assq 'colspan (nth 1 xml-expr)))))
    (setq col-span
	  (cond
	   ((integerp
	       (setq col-span
		     (if (stringp col-span)
			 (string-to-number col-span)
		       col-span)))
	      col-span)
	   ((null col-span) 1)
	   (t (error "colspan invalide"))))  ))

(html2texi-make-simple-markup-handler sub)
(html2texi-make-simple-markup-handler sup)

(html2texi-make-simple-markup-handler tt)

(defvar html2texi-table-fmt-current-ctxt nil)

(defun html2texi-tag-handler-table (xml-expr)
  (let* ((html2texi-table-fmt-current-ctxt (html2texi-table-fmt-ctxt "Table formatting context"))
	 (xml-table-info
	  (vector xml-expr;0: table items
		  nil;1: thead items
		  nil;2: tbody items
		  nil;3: tfoot items
		  0;4: bitmap champ trouv:
					; 1 = plain table (found a tr item not a thead|tbody|tfoot)
					; 2 = thead found
					; 4 = tbody found
					; 8 = tfoot found
		  0;5: bitmap trait
		  1;6: en cours de traitement
		  ))
	 (xml-items (cddr xml-expr)))

    ;; tout d'abord on analyse la table pour trouver le nombre de colonne
    (while (or xml-items (/= (aref xml-table-info 4)  (aref xml-table-info 5)))
      (if xml-items
	  (let ((xml-expr (pop xml-items)))
	    (cond
	     ((and (consp xml-expr)
		   (eq (car xml-expr) 'tr))
	      (when (= (aref xml-table-info 6) 1)
		;; cas d'une table avec les lignes directement sous <table> ... </table>
		(and (/= (logand (aref xml-table-info 4)
				 14) 0)
		     (html2texi-error "Table avec  la fois des lignes directement sous <table> ... </table>\
 et des lignes sous une balise <X>...</X> avec X dans {thead, tbody, tfoot}" xml-expr))
		(aset xml-table-info 4 (logior (aref xml-table-info 4) 1)))
	      ;; plus besoin de chercher une ligne: on compte les colonnes sur la
	      ;; premire ligne trouve
	      (setq xml-items nil)
	      (dolist  (xml-expr (cddr xml-expr))
		(cond
		 ((consp xml-expr)
		  (if (memq  (car xml-expr) '(th td))
		      (progn
			(oset html2texi-table-fmt-current-ctxt :col-count
			      (+ (oref html2texi-table-fmt-current-ctxt :col-count)
				 (html2texi-get-col-span xml-expr)))
			(when (> (oref html2texi-table-fmt-current-ctxt :col-count)
				 (oref html2texi-table-fmt-current-ctxt :col-info-length))
			  (let ((l (make-list (- (oref html2texi-table-fmt-current-ctxt :col-count)
						 (oref html2texi-table-fmt-current-ctxt :col-info-length))
					      '(abs 1))))
			    (setcdr (oref html2texi-table-fmt-current-ctxt :col-info-last) l)
			    (oset html2texi-table-fmt-current-ctxt :col-info-last (last l))
			    (oset html2texi-table-fmt-current-ctxt :col-info-length
				  (oref html2texi-table-fmt-current-ctxt :col-count)))))
		    (html2texi-error "balise inattendu dans une table" xml-expr)))
		 ((stringp xml-expr)
		  (unless (string-match
			   "\\`\\s-*\\(\n\\s-*\\)*\\'"
			   xml-expr)
		    (html2texi-error "Chane inattendue" xml-expr)))
		 (t (html2texi-error "lment inattendu" xml-expr)))))

	     ;; la table est organis en thead/tbody/tfoot
	     ((and (consp xml-expr)
		   (memq (car xml-expr) '(thead tbody tfoot)))
	      (let* ((thead 1) (tbody 2) (tfoot 3) (index (symbol-value (car xml-expr))))
		(and (/= (logand (aref xml-table-info 4)
				 (lsh 1 index)) 0)
		     (html2texi-error "Balise `%s' en double dans la table"
				      (aref xml-table-info 0)
				      (symbol-name (car xml-expr))))
		(aset xml-table-info 4 (logior  (aref xml-table-info 4) (lsh 1 index)))
		(aset xml-table-info index xml-expr)))

	     ;; chane qui n'est pas un blanc au beau milieu de la table...
	     ((stringp xml-expr)
	      (unless (string-match
		       "\\`\\s-*\\(\n\\s-*\\)*\\'"
		       xml-expr)
		(html2texi-error "Chane inattendue" xml-expr)))
	     (t (html2texi-error "lment inattendu" xml-expr))))
	;; xml-items est nil

	;; on marque le champ courant (c.--d. table | thead | tbody | tfoot)
	;; comme ayant t trait
	(aset xml-table-info 5 (logior
				(aref xml-table-info 5)
				(logand
				 (aref xml-table-info 4)
				 (aref xml-table-info 6))))
	;; maintenant on cherche s'il en est un champ restant  parcourir
	(let ((index 0)
	      (to-be-processed (logxor  (aref xml-table-info 4)
					(aref xml-table-info 5))))
	  (while (and (/= to-be-processed 0)
		      (= (logand to-be-processed 1) 0))
	    (setq index (1+ index)
		  to-be-processed (lsh to-be-processed -1)))
	  (when (/= to-be-processed 0)
	    (setq xml-items (cdr-safe (cdr-safe (aref xml-table-info index))))
	    (aset xml-table-info 6 (lsh 1 index))) )))


    ;; maintenant qu'on a fini d'analyser la table, on peut la traiter.
    (if (or (> (oref html2texi-table-fmt-current-ctxt :col-count) 2)
	    (null html2texi-handle-two-columns-table-as-@table))
	(progn
	  (insert "@multitable @columnfractions ")
	  (let ((total-weight
		 (math-reduce-vec
		  (lambda (r x)
		    (+ r
		       (cond
			((eq (car x) 'abs) (cadr x))
			((eq (car x) 'rel)
			 (setcar x 'abs)
			 (setcar (cdr x)
				 (* (cadr x)
				    (oref html2texi-table-fmt-current-ctxt :col-info-length))) (cadr x)))))
		  (cons 'vec (oref html2texi-table-fmt-current-ctxt :col-info)))))
	    (insert
	     (mapconcat
	      (lambda (x)
		(number-to-string
		 (/ (float (cadr x)) (oref html2texi-table-fmt-current-ctxt :col-info-length))))
	      (cdr (oref html2texi-table-fmt-current-ctxt :col-info))
	      " ")))
	  (insert "\n"))
      (insert "@table\n"))

    (dotimes (i 4)
      (when (/= 0 (logand (aref xml-table-info 4) (lsh 1 i)))
	(if (/= i 2)
	    (html2texi-process-xml-expr  (aref xml-table-info i))
	  (oset html2texi-table-fmt-current-ctxt :force-head t)
	  (html2texi-process-xml-expr  (aref xml-table-info 2))
	  (oset html2texi-table-fmt-current-ctxt :force-head nil))))

    (if (or (> (oref html2texi-table-fmt-current-ctxt :col-count) 2)
	    (null html2texi-handle-two-columns-table-as-@table))
	(insert "@end multitable\n")
      (insert "@end table\n"))))


(defun html2texi-tag-handler-tr (xml-expr)
  (oset html2texi-table-fmt-current-ctxt :col-number 0)
  (html2texi-process-xml-expr xml-expr)
  (insert "\n")
  (oset html2texi-table-fmt-current-ctxt :row-number
	(1+ (oref html2texi-table-fmt-current-ctxt :row-number))))

(defun html2texi-tag-handler-th (xml-expr)
  (if   (= 0 (oref html2texi-table-fmt-current-ctxt :col-number))
      (if (and (= 0 (oref html2texi-table-fmt-current-ctxt :row-number))
	       (null html2texi-handle-two-columns-table-as-@table))
	  (progn
	    (oset html2texi-table-fmt-current-ctxt :head-on-row-0 t)
	    (insert "@headitem ")
	    (html2texi-process-xml-expr xml-expr))
	(insert "@item ")
	(html2texi-process-xml-expr xml-expr))
    (when (or (> (oref html2texi-table-fmt-current-ctxt :col-count) 2)
	      (null html2texi-handle-two-columns-table-as-@table))
      (insert "@tab "))
    (html2texi-process-xml-expr xml-expr)
    (unless (and  (= 0 (oref html2texi-table-fmt-current-ctxt :row-number))
		  (oref html2texi-table-fmt-current-ctxt :head-on-row-0))
	(insert "\n")))
  (oset html2texi-table-fmt-current-ctxt :col-number
	(1+ (oref html2texi-table-fmt-current-ctxt :col-number))))

(defun html2texi-tag-handler-td (xml-expr)
  (if  (oref html2texi-table-fmt-current-ctxt :force-head)
      (html2texi-tag-handler-th xml-expr)
    (if (= 0 (oref html2texi-table-fmt-current-ctxt :col-number))
	(insert "@item ")
      (if (or (> (oref html2texi-table-fmt-current-ctxt :col-count) 2)
	      (null html2texi-handle-two-columns-table-as-@table))
	  (insert "@tab ")))
    (html2texi-process-xml-expr xml-expr)
    (insert "\n")
    (oset html2texi-table-fmt-current-ctxt :col-number
	  (1+ (oref html2texi-table-fmt-current-ctxt :col-number)))))

(defun html2texi-cur-dir ()
  "Dtermine le rpertoire courant relativement au rpertoire du
HTML racine de dpart. La valeur retourne se termine par une oblique `/'"
  (let ((cur-dir  (nreverse (split-string (cdar html2texi-directory-stack) "/")))
	(ref-dir  (nreverse (split-string html2texi-directory-ref "/"))))
    (unless (and (string= (car cur-dir) "")
		 (string= (car ref-dir) ""))
      (error "Format invalide de rpertoire"))
    (setq cur-dir (nreverse (cdr cur-dir))
	  ref-dir (cdr ref-dir))
    (if (or (string= (car cur-dir) "")
	    (string-match "\\`[a-z]:" (car cur-dir)))
	;; cur-dir est un chemin absolu
	(progn
	  (setq ref-dir (nreverse ref-dir))
	  (while (and cur-dir ref-dir (string= (car cur-dir) (car ref-dir)))
	    (setq cur-dir (cdr cur-dir)
		  ref-dir (cdr ref-dir)))
	  (while ref-dir
	    (push ".." cur-dir)
	    (setq ref-dir (cdr ref-dir)))
	  (concat (mapconcat 'identity cur-dir "/") "/"))
      (while (and cur-dir
		  (cond
		   ((string= (car cur-dir) "..")
		    (unless ref-dir
		      (error "Chemin invalide"))
		    (setq ref-dir (cdr ref-dir)
			  cur-dir (cdr cur-dir)))
		   ((string= (car cur-dir) ".")
		    (setq cur-dir (cdr cur-dir)))
		   (t nil))))
      (dolist (e cur-dir)
	(push e ref-dir))
      (mapconcat 'identity  (nreverse (cons "" ref-dir)) "/") )))

(defun html2texi-anchor-escape (anchor)
  (let (ret)
    (setq anchor
	  (mapconcat
	   'identity
	   (split-string anchor "-") "--"))
    (mapc
     (lambda (x)
       (if
	(or (and (>= x ?a) (<= x ?z))
	     (and (>= x ?A) (<= x ?Z))
	     (and (>= x ?0) (<= x ?9))
	     (member x '(?_ ?- ?/)))
	   (push (string x) ret)
	 (push (format "-%04x" x) ret)))
     anchor)
    (apply 'concat (nreverse ret))))


(defun html2texi-make-anchor (name &optional escape-function)
  (let* ((anchor (expand-file-name (concat (html2texi-cur-dir) name)))
	 (l-a (length anchor))
	 (l-r (length html2texi-directory-ref))
	 (l (min l-a l-r))
	 (start 0)
	 (i -1))
    (while (and
	    (< (setq i (1+ i)) l)
	    (prog1
		(= (aref anchor i) (aref html2texi-directory-ref i))
	      (and (= (aref anchor i) ?/)
		   (setq start (1+ i))))))
    (setq anchor (list (substring anchor start)))
    (dotimes (i (length (split-string (substring html2texi-directory-ref start))))
      (push "../" anchor))
    (setq anchor (apply 'concat anchor))
    (html2texi-string-escape (funcall (or escape-function 'html2texi-anchor-escape) anchor))))

(defun html2texi-insert-locator (name type &optional escape-function)
  (let ((locator (html2texi-make-anchor name  escape-function)))
    (html2texi-add-locator
     html2texi-locator-list
     (make-instance
      'html2texi-locator-info
      :id locator
      :type type
      :position (point)
      :size (length locator)))
    (insert locator)))

(defun html2texi-simple-markup-handle (xml-expr class-alist preamble postamble)
  "\
Traitement d'un balisage simple en insrant le contenue entre un
prologue et un pilogue. Si la l'attribut `class' a une valeur,
et que cette valeur est trouv dans CLASS-ALIST alors le prologue
et l'pilogue en sont dduit par association, sinon PREAMBLE et
POSTAMBLE sont respectivement utilis."
  (let* ((class-attribute (assq 'class (nth 1 xml-expr)))
	 (pre-post (or (cdr-safe
			(and class-attribute
			     (assoc-string (cdr class-attribute) class-alist)))
		      (list preamble postamble))))
    (insert (car pre-post))
   (html2texi-process-xml-expr xml-expr)
   (insert (cadr pre-post))))

(defun html2texi-tag-handler-body (xml-expr)
  (insert "\n\n")
  (html2texi-process-xml-expr xml-expr)
  (insert "\n\n"))

(html2texi-make-simple-markup-handler code)

(defun html2texi-tag-handler-br (xml-expr)
  (insert "@*\n"))

(defun html2texi-tag-handler-div (xml-expr)
  (insert "\n\n")
  (html2texi-process-xml-expr xml-expr)
  (insert "\n\n"))

(html2texi-make-simple-markup-handler dfn)

(defun html2texi-tag-handler-dl (xml-expr)
  (insert "@table @asis\n")
  (html2texi-process-xml-expr xml-expr)
  (insert "@end table\n"))

(defun html2texi-tag-handler-dt (xml-expr)
  (insert "@item ")
  (html2texi-process-xml-expr xml-expr)
  (insert "\n"))

(defun html2texi-tag-handler-dd (xml-expr)
  (html2texi-process-xml-expr xml-expr)
  (insert "\n"))

(html2texi-make-simple-markup-handler em)

(defun html2texi-tag-handler-frameset (xml-expr)
  (insert "\n@menu\n")
  (html2texi-with-restrictions
   (:frameset)
   (html2texi-process-xml-expr xml-expr))
  (insert "@end menu\n"))

(defun html2texi-tag-handler-frame (xml-expr)
  (unless (html2texi-has-restrictions-p html2texi-convert-restrictions-object :frameset)
    (html2texi-error "<frame> hors d'un <frameset> !" xml-expr))
  (let (url text)
    (dolist (x (cadr xml-expr))
      (cond
       ((eq (car x) 'src)
	(setq url (cdr x)))
       ((eq (car x) 'name)
	(setq text (cdr x)))))
    (when url
      (html2texi-process-url url text :@menu-item))))

(defun html2texi-tag-handler-hr (xml-expr)
  (insert "@c <hr/>\n"))

(html2texi-make-simple-markup-handler kbd)

(defun html2texi-tag-handler-html (xml-expr)
  "\
Traitement de la balise html."
  (let* ((attributes (nth 1 xml-expr))
	 (lang (assq 'lang attributes)))
    (when lang
      	(oset html2texi-document-information-object :language (cdr lang))))
    (html2texi-process-xml-expr xml-expr))

(html2texi-make-simple-markup-handler strong)

(defun html2texi-handle-string (str)
  (let (ret (pos0 0) pos1 (len (length str)))
    (while (and
	    (< pos0 len)
	    (setq pos1 (string-match "[{}@]" str pos0)))
      (push (substring str pos0 pos1) ret)
      (push (concat "@" (match-string-no-properties 0 str)) ret)
      (setq pos0 (1+ pos1)))
    (when (< pos0 len)
      (push (substring str pos0 pos1) ret))
    (apply 'concat (nreverse ret))))

(defun html2texi-generate-or-reuse-log-buffer ()
  (if (buffer-live-p html2texi-reusable-log-buffer)
	(with-current-buffer html2texi-reusable-log-buffer
	 (let ((inhibit-read-only t))
	   (erase-buffer)
	   html2texi-reusable-log-buffer))
    (setq html2texi-reusable-log-buffer
	  (let* ((compilation-error-regexp-alist
		  '(html-to-texinfo-error
		    html-to-texinfo-warning
		    html-to-texinfo-info))
		 (b (generate-new-buffer html2texi-log-buffer-name)))
	    (with-current-buffer b
	      (compilation-mode)
	      b)))))

(defun html2texi-fatal-error (format-str xml-expr &rest args)
  (html2texi-with-log
   (insert (aref html2texi-log-error-names 0) ":"
	   (html2texi-encode-filename-in-log (html2texi-current-file html2texi-files-to-do)) ":"
	   (if (eq xml-expr :html2texi-generic-error)
	       ""
	     (format "%d:" (+ line-number html2texi-line-delta)))
	   (apply
	    'format
	    (concat format-str
		    "\n"
		    (aref html2texi-log-error-names 3)
		    ": <<<-----------\n"
		    (aref html2texi-log-error-names 3)
		    ": xml-expr=%S\n"
		    (aref html2texi-log-error-names 3)
		    ": xml-stack=%S\nInfo: ----------->>>\n")
	    `( ,@args ,xml-expr ,html2texi-xml-stack)) ?\n))
  (apply 'error format-str args))

(defmacro html2texi-with-log (&rest body)
  `(progn
     (setq html2texi-log-buffer (html2texi-generate-or-reuse-log-buffer))
     (display-buffer html2texi-log-buffer)
     (let ((line-number (line-number-at-pos)))
       (with-current-buffer html2texi-log-buffer
	 (let ((inhibit-read-only t))
	   ,@body)))))


(defun html2texi-error (format-str xml-expr &rest args)
  (html2texi-with-log
   (insert (aref html2texi-log-error-names 1) ":"
	   (html2texi-encode-filename-in-log (html2texi-current-file html2texi-files-to-do)) ":"
	   (if (eq xml-expr :html2texi-generic-error)
	       ""
	     (format "%d:" (+ line-number html2texi-line-delta)))
	   " "
	   (apply 'format format-str args) ?\n)))

(defun html2texi-warning (format-str xml-expr &rest args)
  (html2texi-with-log
   (insert (aref html2texi-log-error-names 2) ":"
	   (html2texi-encode-filename-in-log (html2texi-current-file html2texi-files-to-do))
	   ":"
	   (if (eq xml-expr :html2texi-generic-error)
	       ""
	     (format "%d:" (+  line-number html2texi-line-delta)))
	   (apply 'format format-str args) ?\n)))

(defun html2texi-info (format-str xml-expr &rest args)
  (html2texi-with-log
   (insert (aref html2texi-log-error-names 3) ":"
	   (html2texi-encode-filename-in-log (html2texi-current-file html2texi-files-to-do))
	   ":"
	   (if (eq xml-expr :html2texi-generic-error)
	       ""
	     (format "%d:" (+  line-number html2texi-line-delta)))
	   (apply 'format format-str args) ?\n)))

(defun html2texi-decode-url (url)
  "Dcode les `%20' et autres squences hexadcimale"
  (with-temp-buffer
    (insert url)
    (goto-char (point-min))
    (while (re-search-forward "%\\([[:xdigit:]]\\{2\\}\\)" nil t)
      (replace-match (string (math-read-radix
			       (match-string-no-properties 1) 16)) t t))
    (when (eq html2texi-url-encoding :html2texi-utf-8)
      (accents-de-utf-8  (point-min) (point-max)))
    (buffer-substring (point-min) (point-max))))

(defun html2texi-process-url (url xml-expr &optional type)
  (when (html2texi-has-restrictions-p html2texi-convert-restrictions-object :ref)
    (html2texi-error "Rfrence recursive" xml-expr))

  (html2texi-with-restrictions
   (:ref)

   (let* ((parsed-url (url-generic-parse-url (html2texi-decode-url url)))
	  (delimiters (case type
			((:@menu-item)  ["* " ":: " ".\n"])
			(t  ["@ref{" "," "}"])))
	  url-list
	  external-url-p
	  i
	  absolute-file-name
	  locator
	  relative-file-name
	  qualified-locator)
     ;; petit hack parce que url-generic-parse-url ne fait pas compltement le
     ;; boulot
     (when (and
	    (null  (aref parsed-url 1))
	    (setq i (string-match "#" (aref parsed-url 6)))
	    (null (aref parsed-url 7)))
       (aset parsed-url 7 (substring (aref parsed-url 6) (1+ i)))
       (aset parsed-url 6 (substring (aref parsed-url 6) 0 i)))

     ;; Analyse des nom de fichiers
     (if
	 ;; si dans la partie nom de fichier on voit qq chose qui resemble  une requte
	 ;; alors c'est sr que c'est une rfrence externe
	 (setq external-url-p
	       (string-match "\\?[._a-zA-Z][._a-zA-Z0-9]*=" (aref parsed-url 6)))
	 (setq absolute-file-name (aref parsed-url 6))
       (setq absolute-file-name
	     (expand-file-name (aref parsed-url 6)
			       (file-name-directory (html2texi-current-file html2texi-files-to-do)))
	     relative-file-name (file-relative-name absolute-file-name html2texi-directory-ref)))

     (push "@uref{" url-list) ; a peut tre dfait ensuite
     ;; URL
     (if (and (eq (aref parsed-url 0)
		  'cl-struct-url)
	      (null (aref parsed-url 1)))
	 ;; cas o il n'y a pas de protocole
	 (cond
	  ;; on pointe vers un fichier HTML, ce n'est donc pas forcment une URL interne
	  ;; => cas suspect
	  ((and (null external-url-p)
		(member (file-name-extension absolute-file-name)
		   '("html" "htm")))
	   (html2texi-add-file-to-do html2texi-files-to-do absolute-file-name)
	   (setq locator (aref parsed-url 7)
		 qualified-locator (if locator (concat relative-file-name "#" locator) relative-file-name))

	   (if (and (file-exists-p absolute-file-name)
		    (null (file-name-absolute-p relative-file-name)))
	       (setq
		url-list (list
			  (setq qualified-locator (html2texi-make-anchor qualified-locator))
			  (aref delimiters 0)))
	     (push (html2texi-string-escape qualified-locator) url-list)))

	  ;; cas d'une URL interne
	  ((and (string= "" (aref parsed-url 6))
		(setq locator (aref parsed-url 7)))
	   (setq qualified-locator
		 (concat
		  (file-relative-name
		   (html2texi-current-file html2texi-files-to-do)
		   html2texi-directory-ref)
		  "#" locator)
		 url-list
		 (list
		  (setq qualified-locator (html2texi-make-anchor qualified-locator))
		  (aref delimiters 0))))

	  ;; cas d'une URL dont on est sr quelle est externe.
	  (t
	   (push (setq qualified-locator (html2texi-string-escape url)) url-list)))
       ;; cas o il y a un protocol
       (if (and (string= (aref parsed-url 1) "file")
		(null external-url-p)
		(file-exists-p absolute-file-name)
		(null (file-name-absolute-p relative-file-name))
		(member (file-name-extension absolute-file-name) '("html" "htm")))
	   ;; cas d'une URL interne
	   (setq 
	    locator (aref parsed-url 7)
	    qualified-locator (if locator (concat relative-file-name "#" locator) relative-file-name)
	    url-list (list
		      (setq qualified-locator (html2texi-make-anchor qualified-locator))
		      (aref delimiters 0)))
	 (push  (setq qualified-locator (html2texi-string-escape url)) url-list)))
     ;; Text
     (when xml-expr
       (push (aref delimiters 1) url-list)
       (push xml-expr url-list))
     (push (aref delimiters 2) url-list)
     (dolist (v (nreverse url-list))
       (when (eq qualified-locator v)
	 (html2texi-add-locator
	  html2texi-locator-list
	  (make-instance
	   'html2texi-locator-info
	   :id qualified-locator
	   :type type
	   :position (point)
	   :size (length qualified-locator)))
	 (when (eq type :@menu-item)
	   (html2texi-add-menu-reference html2texi-files-to-do 
					 absolute-file-name
					 qualified-locator
					 t)))
       (cond
	((stringp v) (insert v))
	((consp v) (html2texi-process-xml-expr v))) ))))

(defun html2texi-tag-handler-center (xml-expr)
  (let ((start-point (point))
	(start-ln (line-number-at-pos))
	end-mark)
    (html2texi-process-xml-expr xml-expr)
    (when (and (<= (point) (+ start-point html2texi-@center-max-size))
	       (> (point) start-point)
	       ;; test histoire que le code soit  l'preuve du temps: il se
	       ;; pourrait qu'on soit dj centr pour une autre raison.
	       (null (save-excursion
		       (goto-char start-point)
		       (looking-at "\\(\n\\|\\s-\\)*@center\\>")))
	       (let (to-do
		     (ok t)
		     xml-expr
		     (l (cdr-safe (cdr-safe xml-expr))))
		 (while (and ok (or to-do l))
		   (if l
		       (progn
			 (setq xml-expr (pop l))
			 (cond
			  ((stringp xml-expr))
			  ((and (consp xml-expr)
				(memq (car xml-expr) html2texi-allowed-markup-in-@center))
			   (push xml-expr to-do))
			  (t (setq ok nil))))
		     (setq l (cdr-safe (cdr-safe (pop to-do))))))
		 (when ok
		   (setq end-mark (point-marker))
		   (goto-char start-point)
		   (insert "\n@center ")
		   (while (search-forward "\n" end-mark t)
		     (delete-char -1)
		     (insert 32))
		   (goto-char end-mark)
		   (set-marker end-mark nil)))))))

(html2texi-make-simple-markup-handler i)

(defun html2texi-tag-handler-li (xml-expr)
  (insert "\n@item\n")
  (unless (memq (caadr html2texi-xml-stack) '(ol ul))
    (html2texi-fatal-error "<li> tait inattendu." xml-expr ))
  (html2texi-process-xml-expr xml-expr))

(defun html2texi-tag-handler-link (xml-expr)
  )

(defun html2texi-tag-handler-ol (xml-expr)
  (insert "\n@enumerate")
   (html2texi-process-xml-expr xml-expr)
  (insert "\n@end enumerate\n"))

(defun html2texi-tag-handler-p (xml-expr)
  (insert "\n\n")
  (html2texi-process-xml-expr xml-expr)
  (insert "\n\n"))

(defun html2texi-tag-handler-tbody (xml-expr)
  (html2texi-process-xml-expr xml-expr))

(defun html2texi-tag-handler-thead (xml-expr)
  (html2texi-process-xml-expr xml-expr))

(defun html2texi-tag-handler-tfoot (xml-expr)
  (html2texi-process-xml-expr xml-expr))

(defun html2texi-tag-handler-ul (xml-expr)
  (insert "\n@itemize")
  (html2texi-process-xml-expr xml-expr)
  (insert "\n@end itemize\n"))

(defun html2texi-tag-handler-span (xml-expr)
  (insert "@c span: (<span #1>#2</span> => #2.")
  (html2texi-default-handling xml-expr "span: ")
  (html2texi-process-xml-expr xml-expr)
  (insert "@c span: )\n"))

(defun html2texi-tag-handler-style (xml-expr)
  )

(defun html2texi-tag-handler-meta (xml-expr)
  (unless (and (nth 1 html2texi-xml-stack)
	       (eq (caadr html2texi-xml-stack) 'head))
    (html2texi-fatal-error "<meta> inattendu." xml-expr))
  ;; traitement du meta...
  (let* ((attribute-list (nth 1 xml-expr))
	 (http-equiv (assq 'http-equiv attribute-list))
	 (name (assq 'name attribute-list))
	 (content (assq 'content attribute-list)))
    (cond
     ((and (consp name)
	   (consp content)
	   (progn
	     (setq name (cdr name)
		   content (cdr content))
	     (stringp name))
	   (stringp content))
      (cond
       ((string= name "author")
	(oset html2texi-document-information-object :author content))
       ((string= name "language")
	(when (string-match "\\`\\([a-z]\\{2\\}\\(-[A-Z]\\{2\\}\\)?\\)\\'" content)
	  (let ((language (match-string-no-properties 1 content)))
	    (when (= (length language) 5)
	      (aset language 2 ?_))
	    (oset html2texi-document-information-object :language language))))))
     ((and (consp http-equiv)
	   (consp content)
	   (progn (setq http-equiv (cdr http-equiv))
		  (stringp http-equiv))
	   (progn (setq content (cdr content))
		  (stringp content)))
      (setq http-equiv (downcase http-equiv))
      (cond
       ((and (string= http-equiv "content-type")
	     (string-match "charset\\s-*=\\s-*\\([-a-z0-9]+\\)" content))
	(oset html2texi-document-information-object
	      :encoding
	      (match-string-no-properties 1 content)));
       )))))

(defun html2texi-tag-handler-pre (xml-expr)
  (let ((kes html2texi-keep-empty-strings))
    (setq html2texi-keep-empty-strings t)
    (html2texi-process-xml-expr xml-expr)
    (setq html2texi-keep-empty-strings kes)))

(defun hmtl2texi-to-plain-text (xml-expr &rest flags)
  (let (ret anchor)
    (dolist (xml-expr (cddr xml-expr))
      (cond
       ((stringp xml-expr)
	(push xml-expr ret))
       ((consp xml-expr)
	(push xml-expr html2texi-xml-stack)
	(cond
	 ((and (eq (car xml-expr) 'a)
	       (setq anchor (assq 'name (nth 1 xml-expr))))
	  (push (concat
		 (file-relative-name
		  (html2texi-current-file html2texi-files-to-do)
		   html2texi-directory-ref)
		 "#" (cdr anchor))
		html2texi-flushable-anchors) ))
	(let ((str  (hmtl2texi-to-plain-text xml-expr)))
	  (and (null (string= str ""))
	       (push str ret)))
	(pop html2texi-xml-stack))
       (t (html2texi-fatal-error "Expression XML inattendue." xml-expr))))
    (setq  ret (mapconcat 'identity (nreverse ret) " "))
    (if (memq :one-line flags)
	(mapconcat 'identity (split-string ret "\n") " ")
      ret)))

(defun html2texi-tag-handler-title (xml-expr)
  (unless (and  (nth 1 html2texi-xml-stack)
		(eq  (caadr html2texi-xml-stack) 'head))
    (html2texi-fatal-error "<title> inattendu." xml-expr))
  (let ((str  (hmtl2texi-to-plain-text xml-expr :one-line)))
    (setq str (split-string str "\n")
	  str (mapconcat 'identity str " "))
    (unless (string= str "")
      (oset html2texi-document-information-object :title str))))

(defun html2texi-string-escape (str &optional flatten)
  (cond
   ((stringp str)
    (with-temp-buffer
      (insert str)
      (goto-char (point-min))
      (while (re-search-forward "[,@{}]" nil t)
	(cond
	 ((string= (match-string-no-properties 0) ",")
	  (replace-match "@comma{}"))
	 ((member (match-string-no-properties 0) '("@" "{" "}"))
	  (replace-match (concat "@" (match-string-no-properties 0))))))
      (when flatten
	(goto-char (point-min))
	(while (re-search-forward "\n\\(\\s-*\\)" nil t)
	  (replace-match (if (> 0 (length (match-string 1))) " " "") t t)))
      (buffer-substring (point-min) (point-max))))
   ((and (consp str) (car-safe str))
    (cond
     ((eq (car str) 'span)
      (with-temp-buffer
	(insert "@c span: (<span #1>#2</span> => string-escape of #2.")
	(html2texi-default-handling str "span: ")
	(insert (html2texi-string-escape  (nth 2 str) flatten))
	(insert "@c span: )\n")
	(buffer-substring (point-min) (point-max))))
     (t (html2texi-fatal-error "Une chane tait attendue" :html2texi-generic-error str))))
   (t  (html2texi-fatal-error "Une chane tait attendue" :html2texi-generic-error str))))

(defun html2texi-tag-handler-img (xml-expr)
  (let (filename width height alttext extension)
    (dolist (x (cadr xml-expr))
      (cond
       ((eq (car x) 'src)
	(setq filename (cdr x)))
       ((eq (car x) 'alt)
	(setq alttext (html2texi-string-escape (cdr x) t)))))
    (unless filename
      (html2texi-fatal-error "src=... tait attendu" xml-expr))
    (setq filename (html2texi-decode-url filename))
    (setq extension (file-name-extension filename)
	  filename (file-name-sans-extension filename))
    (when (member extension '("png" "jpg" "jpeg" "eps" "txt"))
      (setq extension nil))
    (insert "@image{" (html2texi-make-anchor filename (symbol-function 'identity)))
    (let ((remainder (list width height alttext extension)))
	(while remainder
	  (if (let (non-empty)
		(mapc (lambda (x) (setq non-empty (or non-empty (stringp x))))
		      remainder)
		non-empty)
	      (insert "," (or (pop remainder) "") )
	    (setq remainder nil); rompt la boucle (while remainder...)
	    ))
	(insert "}"))))

(defun html2texi-tag-handler-head (xml-expr)
  (unless html2texi-ignore-head
    (html2texi-process-xml-expr xml-expr)
    (setq html2texi-ignore-head t)))

(defun html2texi-tag-handler-noframes (xml-expr)
  )

(if (boundp 'html2texi-handler-hash-table)
    (makunbound 'html2texi-handler-hash-table))

(defconst html2texi-handler-hash-table
  (let ((ht (make-hash-table)))
    (dolist (v '(a b body center code dfn dl dt dd em i kbd li p
    hr div ol ul pre head meta title frameset frame noframes span
    strong table tbody thead tfoot th tr td h1 h2 h3 h4 h5 h5
    html link br img samp style sup sub tt))
      (puthash v (symbol-function (intern (concat "html2texi-tag-handler-" (symbol-name v))))
	       ht))
    ht)
  "Table de hashage des traitements associs  chaque balise HTML"
  )

(defun html2texi-remove-empty-strings (xml-expr)
  (setq xml-expr (cdr xml-expr))
  (save-match-data
  (while (cdr xml-expr)
    (if (and (stringp (cadr xml-expr))
	     (string-match "\\`\\([\n\r]\\|\\s-\\)*\\'" (cadr xml-expr)))
	(setcdr xml-expr (cddr xml-expr))
      (setq xml-expr (cdr xml-expr))))))

(defun html2texi-process-xml-expr (xml-expr)
  (push xml-expr html2texi-xml-stack)
  (dolist (xml-expr (cddr xml-expr))
    (cond
     ((consp xml-expr)
      (push xml-expr html2texi-xml-stack)
      (let ((handler (gethash (intern (downcase (symbol-name (car xml-expr))))
			      html2texi-handler-hash-table)))
	(if handler
	    (funcall handler  xml-expr)
	  (html2texi-default-handling xml-expr)))
      (pop html2texi-xml-stack))
     ((stringp xml-expr)
      (and (or html2texi-keep-empty-strings
	       (null (string-match "\\`\\([\n\r]\\|\\s-\\)*\\'"  xml-expr)))
	   (insert (html2texi-handle-string xml-expr))))
     (t (html2texi-fatal-error "Expression XML inattendue %S" xml-expr))))
  (pop html2texi-xml-stack))

(defun html2texi-default-handling (xml-expr &optional prompt)
  (let ((str (split-string (prin1-to-string xml-expr) "\n")))
    (dolist (str-line str)
      (insert "\n@c " (or prompt "") str-line)))
    (insert "\n"))

(defun html2texi-process-region (beg end)
  (goto-char end)
  (let ((end (point-marker))
	is-xhtml
	re-do
	xml-expr)

    ;; Suppression de tout ce qui est en dehors des balise <html> ... </html>
    (goto-char beg)
    (setq html2texi-line-delta (line-number-at-pos))
    (setq is-xhtml
	  (looking-at "[ \t\n\r]*<!DOCTYPE[ \t\n\r]+html[ \t\n\r]+PUBLIC[ \t\n\r]+\"-//W3C//DTD XHTML"))
    (unless (re-search-forward "<\\(?:html\\|HTML\\)" end t)
      (html2texi-fatal-error "Balise <html> non trouve" xml-expr))
    (setq html2texi-line-delta (- (line-number-at-pos) (* 2 html2texi-line-delta)))
    (delete-region beg (match-beginning 0))
    (unless (re-search-forward ">" end t)
      (html2texi-fatal-error "Clture de la balise <html> trouve" xml-expr))
    (unless  (re-search-forward "</\\(?:html\\|HTML\\)" end t)
      (html2texi-fatal-error "Balise </html> non trouve trouve" xml-expr))
    (unless (re-search-forward ">" end t)
      (html2texi-fatal-error "Clture de la balise </html> trouve" xml-expr))
    (delete-region (match-end 0) end)

    (or is-xhtml
	(html2texi-make-html-clean-xml beg end))

    (setq xml-expr
	  (condition-case
	      sig
	      (xml-parse-region beg end)
	    (error
	     (if (consp sig)
		 (html2texi-warning "File is XHTML but xml-parser reported error `%S'"
				 :html2texi-generic-error
				 (cdr sig))
	       (html2texi-warning "File is XHTML but xml-parser reported errors"
				  :html2texi-generic-error))
		  (if is-xhtml :html2texi-redo nil)))
	  xml-expr
	  (if (eq xml-expr :html2texi-redo)
	       (progn
		 (html2texi-make-html-clean-xml beg end)
		 (xml-parse-region beg end))
	     xml-expr))

    (delete-region beg end)
    (set-marker end nil)
    xml-expr))

(if t
;; plus partique pour dboguer qu'un vrai tampon temporaire
    (defmacro html2texi-with-temp-buffer (&rest body)
      (let ((cur-buff (make-symbol "cur-buff")))
	`(with-current-buffer
	       (let (( ,cur-buff (get-buffer "*HTML2TEXI Temp*")))
		 (and ,cur-buff (kill-buffer ,cur-buff))
		 (get-buffer-create "*HTML2TEXI Temp*"))
       (erase-buffer)
       ,@body)))
  ;;
  (defmacro html2texi-with-temp-buffer (&rest body)
    `(with-temp-buffer ,@body)))

(defun html2texi-make-texi-buffer (&optional buffer )
  (let* ((start-buffer (or buffer (current-buffer)))
	 xml-expr
	 (start-filename  (or (buffer-file-name start-buffer) (buffer-name)))
	 (start-filename-ext (file-name-extension start-filename))
	 done-links-list
	 texi-buffer)
    (setq html2texi-texi-buffer-name
	  (concat (concat  (file-name-sans-extension   (file-name-nondirectory start-filename))
			   ".texi")))

    (unless (or (member start-filename-ext '("html" "htm"))
		(y-or-n-p (format "le tampon %s n'a pas une extension html, continuer?"
				  start-filename)))
      (html2texi-fatal-error "Fichier `%s' sans extension html" :html2texi-generic-error start-filename))
    (setq texi-buffer (get-buffer-create html2texi-texi-buffer-name))
    (set-buffer texi-buffer)
    (erase-buffer)
    (dolist (v html2texi-texi-buffer-local-variables)
      (set (make-local-variable v) nil))
    (push (cons default-directory "./") html2texi-directory-stack)
    (setq html2texi-directory-ref default-directory)
    (html2texi-add-file-to-do html2texi-files-to-do start-filename)
    (let ((in-menu-table (oref html2texi-files-to-do :in-menu-table)))
      (puthash start-filename 
	       (make-instance 'html2texi-node-info :locator "Top")
	       in-menu-table))
    (while (html2texi-files-has-to-do html2texi-files-to-do)
      (let ((file-name (html2texi-get-next-file-next-to-do html2texi-files-to-do)))
	(if (file-exists-p file-name)
	    (progn
	      (let* ((dir (file-name-as-directory
			   (file-name-directory file-name)))
		     (rel-file-name (file-relative-name file-name html2texi-directory-ref))
		     (rel-dir   (let ((d (file-name-directory rel-file-name )))
				  (if d (file-name-as-directory d)
				    "./"))))
		(push (cons dir rel-dir) html2texi-directory-stack)
		(if (= (oref html2texi-files-to-do :added-file-count) 1)
		    (progn
		      ;; premier fichier trait
		      (insert "@anchor{")
		      (html2texi-insert-locator rel-file-name :@anchor)
		      (insert "}\n"))
		  ;; fichier suivant trait
		  (insert "\n@node ")
		  (html2texi-insert-locator rel-file-name :@node)
		  (insert "\n"))
		(html2texi-with-temp-buffer
		 (insert-file-contents file-name)
		 (accents-de-html)
		 (html2texi-make-html-clean-xml (point-min) (point-max))
		 (setq xml-expr (html2texi-process-region (point-min) (point-max))))
		(unless (eq 'html (caar xml-expr))
		  (html2texi-fatal-error "Rsultat d'analyse XML inattendu" xml-expr))
		(setq xml-expr (car xml-expr))
		(html2texi-process-xml-expr xml-expr)
		(html2texi-files-make-menu-maybe html2texi-files-to-do)
		(pop html2texi-directory-stack)	))
	  (html2texi-warning "Le fichier `%s' n'existe pas!" :html2texi-generic-error file-name))))))

(defun html2texi-insert-doc-info ()
  (let ((author
	 (html2texi-string-escape
	  (if (slot-boundp html2texi-document-information-object :author)
	      (oref  html2texi-document-information-object :author)
	    "AUTHOR")))
	(title 	 (html2texi-string-escape
		  (if (slot-boundp html2texi-document-information-object :title)
		      (oref  html2texi-document-information-object :title)
		    "TITLE")))
	(language (if (slot-boundp html2texi-document-information-object :language)
		       (cons "" (html2texi-string-escape (oref html2texi-document-information-object
							       :language)))
		    (cons "@c " "LANGUAGE")))
	(encoding  (html2texi-string-escape
		  (if (slot-boundp html2texi-document-information-object :encoding)
		      (oref  html2texi-document-information-object :encoding)
		    "iso-8859-1"))))

    (goto-char (point-min))
    (insert "\\input texinfo   @c -*-mode:texinfo; coding:" (downcase encoding) "-*-
@setfilename " (file-name-sans-extension (buffer-name)) ".info
" (car language) "@documentlanguage " (cdr language) "
@documentencoding "  (if (let ((case-fold-search t))
			   (string-match "\\`\\(us\\|utf\\|iso\\)" encoding))
			 (upcase encoding) encoding) "
@copying
This manual is for PROGRAM, version VERSION.

Copyright @copyright{} YEARS COPYRIGHT-OWNER.

@quotation
Permission is granted to ...
@end quotation
@end copying

@titlepage
@title " title "@c NAME-OF-MANUAL-WHEN-PRINTED
@c @subtitle SUBTITLE-IF-ANY
@c @subtitle SECOND-SUBTITLE
@author " author "

@c  The following two commands
@c  start the copyright page.
@page
@vskip 0pt plus 1filll
@insertcopying

Published by ...
@end titlepage

@c So the toc is printed at the start.
@contents

@ifnottex
@node Top
@top TITLE

This manual is for PROGRAM, version VERSION.
@end ifnottex
")
    (oset html2texi-document-information-object :header-size (- (point) (point-min)))
    (goto-char (point-max))
    (insert "
@bye") ))

(defun html2texi-beautify-locator (locator)
  (with-temp-buffer
    (insert locator)
    (goto-char (point-min))
    (when (re-search-forward "-002ehtml?\\(-0023\\)?" nil t)
      (if
	  (match-string 1)
	  (replace-match "_" t t)
	(replace-match "" t t)))
    (goto-char (point-min))
    (while (re-search-forward "-[[:xdigit:]]\\{4\\}" nil t)
      (replace-match "-" t t))
    (buffer-substring (point-min) (point-max))))

(defun html2texi-post-process ()
  "Embellit les identificateur de localisation selon
`html2texi-beautify-locators'. Puis pemplace les double lignes
vides en ligne vides simples."
  ;; embellissement des indentificateurs de localisateur
  (when html2texi-beautify-locators
    (html2texi-info "Embellissement des localisateurs..." :html2texi-generic-error )
    (let ((offset (oref html2texi-document-information-object :header-size)))
      (doli2->-dolist (v html2texi-locator-list)
		      (oset v :position (+ offset (oref v :position)))))
    (let ((dummy-tail (html2texi-locator-info "Factice"))
	  key-val-pairs
	  (table (oref html2texi-locator-list :hash-table)))
      (doli2-add-last html2texi-locator-list dummy-tail)
      (maphash
       #'(lambda (key val)
	   (push (cons key val) key-val-pairs))
       table)
      (dolist (key-val key-val-pairs)
	(let* ((key (car key-val))
	      (val (cdr key-val))
	      (beautiful-key (html2texi-beautify-locator key))
	      delta-pos-inc delta-pos locator-list
	      (key-length (length key))
	      cur next)
	  (if (string= beautiful-key key)
	      (puthash key nil table)
	    (setq delta-pos-inc (- (length beautiful-key) key-length)
		  delta-pos 0
		  locator-list (reverse (cons dummy-tail val))
		       next (pop locator-list))
	    (save-excursion
	      (while (null (eq next dummy-tail))
		(setq cur next
		      next (pop locator-list))
		(goto-char (+ (oref cur :position) delta-pos))
		(let* ((from (point))
		       (to (+ from key-length)))
		  (unless (string= (buffer-substring-no-properties from to) key)
		    (html2texi-fatal-error "Tampon corrompu: trouv `%s' et attendait `%s'"
					   :html2texi-generic-error
					   (buffer-substring-no-properties from to)
					   key))
		  (delete-region from to))
		(insert beautiful-key)
		(setq delta-pos (+ delta-pos delta-pos-inc))
		(doli2-x>x-dolist
		 (v cur next)
		 (oset v :position (+ (oref v :position) delta-pos)))
		(doli2-remove html2texi-locator-list cur)))
	    (remhash key table)
	    (puthash beautiful-key t table))))
      (doli2-remove-last  html2texi-locator-list)))

  ;; suppression des doubles lignes en trop
  (goto-char (point-min))
  (while (re-search-forward "\\(^[ \t]*\n\\)\\{2,\\}" nil t)
    (replace-match "\n"))
  (normal-mode))

(defun html2texi-save-texi-buffer-maybe ()
  "Sauvegarde le tampon avec le fichier Texinfo."
  (if
      (buffer-file-name) ; le tampon visite dj un fichier
      (basic-save-buffer)
   (write-file (buffer-name) html2texi-save-texi-buffer-confirm-p)))

;;;###autoload
(defun html2texi ()
  (interactive)
  (let (html2texi-keep-empty-strings
	html2texi-xml-stack
	html2texi-texi-buffer-name
	(html2texi-line-delta 0)
	html2texi-ignore-head
	html2texi-directory-stack
	html2texi-flushable-anchors
	html2texi-directory-ref
	(html2texi-locator-list (html2texi-locator-listing " embellir"))
	(html2texi-files-to-do (html2texi-files-to-do-listing " traiter"))
	(html2texi-document-information-object (html2texi-document-information "Info doc"))
	(html2texi-log-buffer (and html2texi-reuse-log-buffer html2texi-log-buffer))
	(html2texi-convert-restrictions-object (html2texi-convert-restrictions "Restrictions de conversion")))
  (html2texi-make-texi-buffer)
  (html2texi-insert-doc-info)
  (html2texi-post-process)
  (html2texi-info "Fin de la conversion en HTML!" :html2texi-generic-error )
  (html2texi-save-texi-buffer-maybe)))

;; Log compilation mode stuff
;;-----------------------------------------------------------------------------
(defun html2texi-decode-filename-in-log (filename)
  (save-match-data
    (cond
     ((string-match "\\`[^ \t\n><:\"][^ \t\n><:]*\\'" filename)
      filename)
     ((string-match "\\`\"\\(?:[^\\\"]\\|\\\\[^\\\"]\\)+\"\\'" filename)
    (substring filename 1 -1))
     ((string-match "\\`\"\\(?:[^\\\"]\\|\\\\[\\\"]\\)+\"\\'" filename)
      (setq filename (substring filename 1 -1))
      (let ((pos 0) next-pos)
	(while (and 
		(< pos (length filename))
		(setq next-pos (string-match "\\\\[\\\"]" filename pos)))
	  (setq filename (concat (substring filename 0 next-pos)
				 (substring filename (1+ next-pos)))
		pos   (1+ next-pos))))
      filename)
     (t 
      ;; on ne devrait jamais arriver ici, sauf bogue de ma part
      (error "Nom de fichier inattendu")))))

(defun html2texi-encode-filename-in-log (filename)
  (cond 
   ;; cas  la Linux: le nom ne contient aucun caractre spcial
   ((string-match "\\`[^ \t\n><:\"][^ \t\n><:]*\\'" filename)
    ;; on passe le nom en clair
    filename)
   ;; cas lger  la MSWindows: ne contient pas de ", ni ne finit par \
   ((null 
     (or
      (string-match "\"" filename)
      (string= (substring filename -1) "\\")))
    ;; on encapsule entre guillemets anglais
    (concat "\"" filename "\""))
   ;; cas lourd  la MSWindows:
   (t
    ;; on chappe tous les \ et " par un \, puis on encapsule entre "
    (let ((pos 0) next-pos)
      (while (and
	      (< pos (length filename))
	      (setq next-pos (string-match "[\\\"]" filename pos)))
	(setq filename (concat (substring filename 0 next-pos)
			       "\\"
			       (substring filename next-pos))
	      pos (+ 2 next-pos))))
    (concat "\"" filename "\""))    ))

(defun html2texi-get-filename-in-log ()
  (list (html2texi-decode-filename-in-log (match-string 1))))

(defun html2texi-define-error-regexps ()
  (dolist (v `(
	       (html-to-texinfo-error
		.
		(
		 ,(concat "^" (regexp-opt (list (aref html2texi-log-error-names 0)
						(aref html2texi-log-error-names 1)))
			  ":\\(" html2texi-filepath-re
			  "\\):\\(?:\\([0-9]+\\):\\)?.*$")
		 html2texi-get-filename-in-log; File
		 2; Line
		 0; Error
		 ))
	       (html-to-texinfo-warning
		.
		(
		 ,(concat "^" (aref html2texi-log-error-names 2)
			  ":\\(" html2texi-filepath-re
			  "\\):\\(?:\\([0-9]+\\):\\)?.*$")
		 html2texi-get-filename-in-log; File
		 2; Line
		 1; Warning
		 ))
	       (html-to-texinfo-info
		.
		(
		 ,(concat "^" (aref html2texi-log-error-names 3)
			  ":")
		 nil; File
		 nil; Line
		 2; Warning
		 ))))
    (add-to-list 'compilation-error-regexp-alist (car v))
    (let ((cell (or (assq (car v) compilation-error-regexp-alist-alist)
		    (car (push (cons  (car v) nil) compilation-error-regexp-alist-alist)))))
      (setcdr cell (cdr v))  )))
(html2texi-define-error-regexps)

;;; html-to-texinfo.el ends here
